function out = IdealObserver_TP_uncoupledCP(in)
% Wrapper function.
% This observer computes the Bayes-optimal (HMM) inference of transition
% probabilities underlying a binary sequence of observation, with the
% assumption that those transition probabilities undergo rare, sudden 
% changes, and that change points are uncoupled: only one of the two 
% transition probabilities is changed at a time, and each time a change 
% point occurs, the identity of the transition probability which changes 
% is random.
%
% This inference actually amounts to computing two inference in parallel 
% and independently from one another, for each transition type. This 
% inference therefore amounts to performing parallel, independent HMM for 
% each transition type. This HMM algorithm is part of online toolbox:
% https://github.com/florentmeyniel/MinimalTransitionProbsModel
%
% Florent Meyniel 2019.

% Add a few private functions
currdir = pwd;
IOdir = which('IdealObserver.m');
filepath = fileparts(IOdir);
cd([filepath, '/private']);

% REFORMAT THE SEQUENCE
% =====================
% Re-format the input sequence into two sequences, each corresponding to
% the observation that follows a specific item. The result is a set of two 
% sequences, one for transitions from observation "1", and the other for 
% transitions from observation "2".
% Trials that are empty (the corresponding transition type is not observed)
% are filled with NaN.
s_from1 = nan(1, numel(in.s));
s_from2 = nan(1, numel(in.s));

for iObs = 2:numel(in.s)
    if in.s(iObs-1) == 1
        s_from1(iObs) = in.s(iObs);
    elseif in.s(iObs-1) == 2
        s_from2(iObs) = in.s(iObs);
    end
end

% RUN THE STANDARD HMM OBSERVER FOR JUMPING FREQUENCY ON EACH SEQUENCE
% ====================================================================
out_perTrans = cell(1, 2);
for iSeq = 1:2
    if iSeq == 1
        s = s_from1;
    else
        s = s_from2;
    end
    
    % specify options
    in_Fhmm.s            = s;                   % sequence
    in_Fhmm.learned      = 'frequency';         % estimate item frequency
    in_Fhmm.jump         = 1;                   % estimate with jumps
    in_Fhmm.mode         = 'HMM';               % use the HMM (not sampling) algorithm
    in_Fhmm.opt.pJ       = in.opt.pJ/2;         % a priori probability that a jump occur at each outcome
    n                    = 100;                 % resolution of the univariate probability grid
    in_Fhmm.opt.pgrid    = in.opt.pgrid;        % estimation probability grid
    n                    = numel(in.opt.pgrid); %
    in_Fhmm.opt.Alpha0   = ones(n,1)/n;         % uniform prior on transition probabilities
    in_Fhmm.verbose      = in.verbose;          % to check that no default values are used.
    in_Fhmm.opt.ReturnDistUpdate = in.opt.ReturnDistUpdate;           % don't return distribution update
    
    % Compute the observer
    out_perTrans{iSeq} = IdealObserver(in_Fhmm);
end

% COMBINE THE TWO ESTIMATES AND FILL THE OUTPUT STRUCTURE
% =======================================================
% get full distribution
out.p1g2_dist = out_perTrans{2}.p1_dist;
out.p2g1_dist = flipud(out_perTrans{1}.p1_dist); % change p(1) into p(2)

% get joint distribution
out.p1g2xp2g1 = nan(n, n, numel(in.s));
for iObs = 1:numel(in.s)
    out.p1g2xp2g1(:, :, iObs) = out.p1g2_dist(:, iObs) * out.p2g1_dist(:, iObs)';
end

% get "posterior" prediction (= mean) and confidence in this prediction
% given each transition type
out.p1g2_mean  = dist_mean(out.p1g2_dist', in.opt.pgrid)';
out.p1g2_sd    = dist_sd(  out.p1g2_dist', in.opt.pgrid)';
out.p2g1_mean  = dist_mean(out.p2g1_dist', in.opt.pgrid)';
out.p2g1_sd    = dist_sd(  out.p2g1_dist', in.opt.pgrid)';

% get prediction and confidence given the current observation
out.p1_mean    = GetConditionalValue(out.p1g2_mean, 1-out.p2g1_mean, in.s);
out.p1_sd      = GetConditionalValue(out.p1g2_sd,   out.p2g1_sd,     in.s);

% get surprise given these predictions and the actual outcomes
out.surprise   = ComputeSurprise(out.p1_mean, in.s);

% get update of the posterior distribution (KL divergence)
% NB: add the prior as 1st estimate.
if in.opt.ReturnDistUpdate
    out.distUpdate  = DistributionUpdateMarkov(cat(3, in.opt.Alpha0 * in.opt.Alpha0', out.p1g2xp2g1));
else
    out.distUpdate  = [];
end

% go back to initial directory
cd(currdir)

end

% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % SUBFUNCTIONS % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %

% Select the value conditional on the current observation
function val = GetConditionalValue(val_g2, val_g1, s)
seqL = numel(s);
val = nan(1, seqL);
for k = 1:seqL
    if s(k) == 1
        val(k) = val_g1(k);
    else
        val(k) = val_g2(k);
    end
end
end

% Compute Shannon surprise given predictions and actual outcomes.
function surp = ComputeSurprise(p1_mean, s)
seqL = numel(s);
surp = nan(1, seqL);
surp(1) = -log2(0.5);
for k = 2:seqL
    if s(k) == 1
        surp(k) = -log2(p1_mean(k-1));      % likelihood of s(k)=1 given s(1:k-1)
    else
        surp(k) = -log2(1-p1_mean(k-1));    % likelihood of s(k)=2 given s(1:k-1)
    end
end
end

% Return the KL divergence between two consecutive 2D distributions
function KLdiv = DistributionUpdateMarkov(Fdist)
% Initialize variables
seqL  = size(Fdist, 3);
KLdiv = nan(1, seqL);

% Convert 0 into eps (the closest value to 0 given Matlab largest
% double-precision) to compute the KL divergence without the numeric
% problem of 0*log(0), which occurs when absolute continuity does not hold.
% NB: These 0s (if any) are on the edges of the distribution. An
% alternative is to crop the distribution and discard the edge values (for
% a 2D distribution, this corresponds to discarding the first and last rows
% and the first and last columns). Both options, cropping or converting 0s
% to eps, asymptotically converge when the resolution of the grid increases.
Fdist_corr = Fdist;
Fdist_corr(Fdist == 0) = eps; % approximate 0 value with Matlab largest double-precision

% Compute KL divergence across successive posterior estimates
for k = 2:seqL
    % Posterior distribution given stimulus 1 to k-1
    dist_prev = squeeze(Fdist_corr(:, :, k-1));
    
    % Posterior distribution given stimulus 1 to k
    dist_curr = squeeze(Fdist_corr(:, :, k));
    
    % KL divergence, i.e. distribution Update (note that the KL is not symmetric
    % so the order matters!! KL(Post || Prior) quantify how much
    % information is gained when one revises Prior into Post)
    KLdiv(k) = dist_KLdiv(dist_curr(:)', dist_prev(:)');
end

% Note that the first element in Fdist is actually the prior. Therefore, 
% currently, KLdiv(1) is nan, KLdiv(2) is the update between prior and 1st
% posterior, KLdiv(3) is the update between 1st and 2nd posterior, etc. We
% just want to get rid of the 1st nan and align KLdiv(k) on the update
% induced by the k-th observation.
KLdiv = KLdiv(2:end);
end
