% This Matlab script accompanies the study by Heilbron and Meyniel, 
% "Confidence resets reveal hierarchical adaptive learning in humans".
% Publised in Plos Computational Biology, 2019.
% This study uses several ideal observers. The core codes for running those
% models are provided as a Matlab toolbox available on GitHub:
% https://github.com/florentmeyniel/MinimalTransitionProbsModel 
%
% To use this script, make sure first to download this toolbox, and to
% include in your Matlab's paths two of its sub-directories: 
%   /IdealObserversCode
%   /Tools
%
% Each of the cell of this script shows an example use of the Ideal
% Observer models used in our study.
%
% Florent Meyniel 2019.

%% Generate a binary sequence with changing transition probabilities

% Parameters of the sequence
p_c     = 1/75; % probability of change point on each trial
p_min   = 0.1;  % minimum transition probability
p_max   = 0.9;  % maximum transition probability
L       = 380;  % length of the sequence

% Randomly allocate change points, with the constraint that they are not at
% the very beginning or very end. Those changes points are the same for the
% two transition probabilities.
p_c_position = [false, rand(1, L-2) < p_c, false];
chunk_length = diff(find([true, p_c_position(2:L), true]));

% Randomly sample the transition probabilities in the range [p_min p_max].
% The two possible observations are denoted "1" and "2". The probability of
% observing a "1" after a "2", p(1|2), is denoted p1g2 and the probability 
% of observing a "2" after a "1", p(2|1), is denoted p1g2.
p1g2 = rand(1, sum(p_c_position)+1)*(p_max-p_min)+p_min;
p2g1 = rand(1, sum(p_c_position)+1)*(p_max-p_min)+p_min;

% Generate a binary, random sequence following those changing transition
% probabilities.
[s, gen_p1, gen_p1g2, gen_p2g1 ] = GenRandSeq(chunk_length, [p1g2', p2g1']);

% Note:
% In the paper, the sequence generation is slightly more constrained:
%   - we exclude sequences with overly long stable periods
%   - we impose a constraint about the change in probabilities, which
%   should be large enough for at least one of the two transition
%   probabilities.

%% Hierarchical model with coupled change points

% Options corresponding to this model:
clear('in');
in.s            = s;                % sequence
in.learned      = 'transition';     % it estimates transition probabilities
in.jump         = 1;                % it assumes jumps (i.e. volatility)
in.mode         = 'HMM';            % use the HMM algorithm
in.opt.pJ       = p_c;              % a priori volatility
n               = 50;               % resolution of the univariate probability grid
in.opt.pgrid    = linspace(0,1,n);  % estimation probability grid
in.opt.Alpha0   = ones(n)/(n^2);    % uniform prior on transition probabilities
in.verbose      = 0;                % Don't report checks.

% Run the Ideal observer
out = IdealObserver(in);

% Plot the result of the inference
figure; set(gcf, 'Color', [1 1 1])
set(gcf, 'Name', 'Hierarchical model for coupled change points')
PlotResults

% Notes:
%   - The a priori volatility in.opt.pJ can be set to a value fitted onto 
%     the subject's reports.
%   - In the paper, we also simulate a hierarchical model that assumes that
%     observations are generated based on the stimulus frequency, p(1),
%     rather than transition probabilities. To simulate such a model,
%     change the following options:
%       >> in.learned      = 'frequency';
%       >> in.opt.Alpha0   = ones(n,1)/n;
%     This model will not return the posterior p(1|2) and p(2|1), but
%     simply p(1).

%% Hierarchical model with uncoupled change points

% If the model assumes that change points are not coupled across the two
% transition probabilities, then those transition probabilities follow
% independent, parallel streams and they can be estimated fully
% independently from one another.
% Technically, this model amounts to running two ideal observers, one per
% transition probability, independently. We provide a separate wrapper
% function to achieve this computation.

% Options corresponding to this model:
clear('in');
in.s                    = s;                % sequence
in.opt.pJ               = p_c;              % a priori volatility
n                       = 50;               % resolution of the univariate probability grid
in.opt.pgrid            = linspace(0,1,n);  % estimation probability grid
in.opt.Alpha0           = ones(n)/(n^2);    % uniform prior on transition probabilities
in.opt.ReturnDistUpdate = 1;                % do not return distribution update
in.verbose              = 0;                % Don't report checks.

% Run the Ideal observer
out = IdealObserver_TP_uncoupledCP(in);

% Plot the result of the inference
figure; set(gcf, 'Color', [1 1 1])
set(gcf, 'Name', 'Hierarchical model for uncoupled change points')
PlotResults

%% Flat model

% Options corresponding to this model:
clear('in');
in.s                = s;                % sequence
in.learned          = 'transition';     % it estimates transition probabilities
in.jump             = 0;                % no monitoring of change points (= a flat model)
in.opt.MemParam     = {'Decay', 14.4};  % Parameter of the leaky integration
n                   = 50;               % resolution of the univariate probability grid
in.opt.pgrid        = linspace(0,1,n);  % grid to return full distributions
in.opt.ReturnDist   = 1;                % Return full posterior distributions
in.opt.priorp1g2    = [1 1];            % Parameters of priori beta distribution for p(1|2)
in.opt.priorp2g1    = [1 1];            % Parameters of priori beta distribution for p(2|1)
in.verbose          = 0;                % Don't report checks.

% Run the Ideal observer
out = IdealObserver(in);

% Plot the result of the inference
figure; set(gcf, 'Color', [1 1 1])
set(gcf, 'Name', 'Flat model')
PlotResults

% Notes:
%   - The leak parameter can be set to a value fitted onto the subject's 
%     reports. This parameter is called omega in the paper.
%   - In the paper, we also simulate a flat model that assumes that
%     observations are generated based on the stimulus frequency, p(1),
%     rather than transition probabilities. To simulate such a model,
%     change the following options:
%       >> in.learned      = 'frequency';
%       >> in.opt.priorp1  = [1 1];
%     This model will not return the posterior p(1|2) and p(2|1), but
%     simply p(1).
%   - In the paper, we also simulate the flat model with a different prior,
%     [0 0] instead of [1 1], which corresponds to 0 (rather than 1) prior
%     count of either item. To simulate such a model, simply change the
%     priors:
%       >> in.opt.priorp1   = [0 0];
%       >> in.opt.priorp1g2 = [0 0];
%       >> in.opt.priorp2g1 = [0 0];
